# Copyright 2025 CVS Health and/or one of its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Any, Dict, List, Optional, Tuple
from rich.progress import Progress
from langchain_core.language_models.chat_models import BaseChatModel
from uqlm.scorers.longform.baseclass.uncertainty import LongFormUQ
from uqlm.utils.results import UQResult
from uqlm.longform.graph import GraphScorer, ClaimMerger

GRAPH_SCORERS = ["degree_centrality", "betweenness_centrality", "closeness_centrality", "page_rank", "laplacian_centrality", "harmonic_centrality"]


class LongTextGraph(LongFormUQ):
    def __init__(
        self,
        llm: Optional[BaseChatModel] = None,
        scorers: Optional[List[str]] = None,
        aggregation: str = "mean",
        response_refinement: bool = False,
        claim_decomposition_llm: Optional[BaseChatModel] = None,
        claim_filtering_scorer: Optional[str] = None,
        device: Any = None,
        nli_model_name: str = "microsoft/deberta-large-mnli",
        system_prompt: str = "You are a helpful assistant.",
        max_calls_per_min: Optional[int] = None,
        sampling_temperature: float = 1.0,
        use_n_param: bool = False,
        max_length: int = 2000,
    ) -> None:
        """
        Class for Long-text Uncertainty Quantification (LUQ) scorers.
        Parameters
        ----------
        llm : langchain `BaseChatModel`, default=None
            A langchain llm `BaseChatModel`. User is responsible for specifying temperature and other
            relevant parameters to the constructor of their `llm` object.
        scorers : List[str], default=None
            Specifies which graph-based scorers to include. Must be subset of ["degree_centrality", "betweenness_centrality",
            "closeness_centrality", "page_rank", "laplacian_centrality", "harmonic_centrality"]. If None, defaults to ["closeness_centrality"].
        granularity : str, default="claim"
            Specifies whether to decompose and score at claim or sentence level granularity. Must be either "claim" or "sentence"
        aggregation : str, default="mean"
            Specifies how to aggregate claim/sentence-level scores to response-level scores. Must be one of 'min' or 'mean'.
        response_refinement : bool, default=False
            Specifies whether to refine responses with uncertainty-aware decoding. This approach removes claims with confidence
            scores below the response_refinement_threshold and uses the claim_decomposition_llm to reconstruct the response from
            the retained claims. Only available for claim-level granularity. For more details, refer to
            Jiang et al., 2024: https://arxiv.org/abs/2410.20783
        claim_filtering_scorer : Optional[str], default=None
            specifies which scorer to use to filter claims if response_refinement is True. If not provided, defaults to the first
            element of self.scorers.
        claim_decomposition_llm : langchain `BaseChatModel`, default=None
            A langchain llm `BaseChatModel` to be used for decomposing responses into individual claims. Also used for claim refinement.
            If granularity="claim" and claim_decomposition_llm is None, the provided `llm` will be used for claim decomposition.
        device: str or torch.device input or torch.device object, default="cpu"
            Specifies the device that NLI model use for prediction. If None, detects and returns the best available PyTorch device.
            Prioritizes CUDA (NVIDIA GPU), then MPS (macOS), then CPU.
        nli_model_name : str, default="microsoft/deberta-large-mnli"
            Specifies which NLI model to use. Must be acceptable input to AutoTokenizer.from_pretrained() and
            AutoModelForSequenceClassification.from_pretrained()
        system_prompt : str or None, default="You are a helpful assistant."
            Optional argument for user to provide custom system prompt
        max_calls_per_min : int, default=None
            Specifies how many api calls to make per minute to avoid a rate limit error. By default, no
            limit is specified.
        sampling_temperature : float, default=1.0
            The 'temperature' parameter for llm model to generate sampled LLM responses. Must be greater than 0.
        use_n_param : bool, default=False
            Specifies whether to use `n` parameter for `BaseChatModel`. Not compatible with all
            `BaseChatModel` classes. If used, it speeds up the generation process substantially when num_responses > 1.
        max_length : int, default=2000
            Specifies the maximum allowed string length. Responses longer than this value will be truncated to
            avoid OutOfMemoryError
        """
        self.scorers = ["closeness_centrality"] if not scorers else scorers
        super().__init__(llm=llm, aggregation=aggregation, scorers=self.scorers, response_refinement=response_refinement, claim_filtering_scorer=claim_filtering_scorer, claim_decomposition_llm=claim_decomposition_llm, device=device, system_prompt=system_prompt, max_calls_per_min=max_calls_per_min, use_n_param=use_n_param)
        self.nli_model_name = nli_model_name
        self.max_length = max_length
        self.sampling_temperature = sampling_temperature
        self.graph_scorer = GraphScorer(nli_model_name=nli_model_name, max_length=max_length, device=device)
        self.claim_merger = ClaimMerger(claim_merging_llm=self.decomposer.claim_decomposition_llm)
        self.prompts = None
        self.responses = None
        self.claim_sets = None
        self.sampled_responses = None
        self.sampled_claim_sets = None
        self.num_responses = None
        self.uad_result = {}

    async def generate_and_score(self, prompts: List[str], num_responses: int = 5, response_refinement_threshold: float = 1 / 3, show_progress_bars: Optional[bool] = True) -> UQResult:
        """
        Generate LLM responses, sampled LLM (candidate) responses, and compute confidence scores with specified scorers for the provided prompts.
        Parameters
        ----------
        prompts : list of str
            A list of input prompts for the model.
        num_responses : int, default=5
            The number of sampled responses used to compute consistency.
        response_refinement_threshold : float, default=1/3
            Threshold for uncertainty-aware filtering. Claims with confidence scores below this threshold are dropped from the
            refined response. Only used if response_refinement is True.
        show_progress_bars : bool, default=True
            If True, displays progress bars while generating and scoring responses
        Returns
        -------
        UQResult
            UQResult containing data (prompts, responses, and scores) and metadata
        """
        self.prompts = prompts
        self.num_responses = num_responses

        self._construct_progress_bar(show_progress_bars)
        self._display_generation_header(show_progress_bars)

        responses = await self.generate_original_responses(prompts=prompts, progress_bar=self.progress_bar)
        sampled_responses = await self.generate_candidate_responses(prompts=prompts, progress_bar=self.progress_bar, num_responses=self.num_responses)
        result = await self.score(responses=responses, sampled_responses=sampled_responses, response_refinement_threshold=response_refinement_threshold, show_progress_bars=show_progress_bars)
        return result

    async def score(self, responses: List[str], sampled_responses: List[List[str]], response_refinement_threshold: float = 1 / 3, show_progress_bars: Optional[bool] = True) -> UQResult:
        """
        Compute confidence scores with specified scorers on provided LLM responses. Should only be used if responses and sampled responses
        are already generated. Otherwise, use `generate_and_score`.
        Parameters
        ----------
        responses : list of str, default=None
            A list of model responses for the prompts.
        sampled_responses : list of list of str, default=None
            A list of lists of sampled LLM responses for each prompt. These will be used to compute consistency scores by comparing to
            the corresponding response from `responses`.
        response_refinement_threshold : float, default=1/3
            Threshold for uncertainty-aware filtering. Claims with confidence scores below this threshold are dropped from the
            refined response. Only used if response_refinement is True.
        show_progress_bars : bool, default=True
            If True, displays a progress bar while scoring responses
        Returns
        -------
        UQResult
            UQResult containing data (responses and scores) and metadata
        """
        self.responses = responses
        self.sampled_responses = sampled_responses
        self.num_responses = len(sampled_responses[0])
        self._construct_progress_bar(show_progress_bars)

        await self._decompose_responses(show_progress_bars)
        await self._decompose_candidate_responses(show_progress_bars)
        await self._merge_claims(show_progress_bars)

        self._display_scoring_header(show_progress_bars)

        all_responses = [[r] + sr for r, sr in zip(self.responses, self.sampled_responses)]

        original_claim_scores, master_claim_scores, graph_score_result = self._score_from_decomposed(original_claim_sets=self.claim_sets, master_claim_sets=self.master_claim_sets, response_sets=all_responses, progress_bar=self.progress_bar)

        if self.response_refinement:
            self.claim_scores = master_claim_scores
            self.uad_result = await self.uncertainty_aware_decode(claim_sets=self.master_claim_sets, claim_scores=self.claim_scores[self.uad_scorer], response_refinement_threshold=response_refinement_threshold, show_progress_bars=show_progress_bars)
        self._stop_progress_bar()
        self.progress_bar = None

        self.scores_dict["claims_data"] = self._unpack_claims_data(graph_score_result)
        if "removed" in self.uad_result:
            del self.uad_result["removed"]

        return self._construct_result()

    def _score_from_decomposed(self, original_claim_sets: List[List[str]], master_claim_sets: List[List[str]], response_sets: List[List[str]], progress_bar: Optional[Progress] = None) -> Tuple[Any, Any, Any]:
        """
        Compute confidence scores with specified scorers on provided LLM responses. Should only be used if responses and sampled responses
        are already generated. Otherwise, use `generate_and_score`.
        Parameters
        ----------
        claim_sets : list of list of strings
            List of original responses decomposed into lists of either claims or sentences
        master_claim_sets : list of list of strings
            Candidate responses to be compared to the decomposed original responses
        response_sets : list of list of strings
            Decomposed responses to be compared to the decomposed original responses
        Returns
        -------
        UQResult
            UQResult containing data (responses and scores) and metadata
        """
        graph_score_result = self.graph_scorer.evaluate(original_claim_sets=self.claim_sets, master_claim_sets=self.master_claim_sets, response_sets=response_sets, progress_bar=progress_bar)
        original_claim_scores, master_claim_scores = self._unpack_results(graph_score_result)
        return original_claim_scores, master_claim_scores, graph_score_result

    def _construct_result(self) -> Any:
        """Constructs UQResult object"""
        data = {"responses": self.responses, "sampled_responses": self.sampled_responses}
        if self.prompts:
            data["prompts"] = self.prompts
        data.update(self.scores_dict)
        data.update(self.uad_result)
        result = {"data": data, "metadata": {"aggregation": self.aggregation, "temperature": None if not self.llm else self.llm.temperature, "sampling_temperature": None if not self.sampling_temperature else self.sampling_temperature, "num_responses": self.num_responses, "response_refinement_threshold": self.response_refinement_threshold}}
        return UQResult(result)

    async def _merge_claims(self, show_progress_bars) -> None:
        self.master_claim_sets = await self.claim_merger.merge_claims(original_claim_sets=self.claim_sets, sampled_claim_sets=self.sampled_claim_sets, progress_bar=self.progress_bar if show_progress_bars else None)

    def _unpack_results(self, result: List[List[Any]]) -> Any:
        original_claim_scores = {k: [] for k in self.scorers}
        master_claim_scores = {k: [] for k in self.scorers}
        self.scores_dict = {k: [] for k in self.scorers}
        for scorer in self.scorers:
            for i in range(len(result)):
                score_list_i = []
                master_score_list_i = []
                for j in range(len(result[i])):
                    master_score_list_i.append(result[i][j].scores[scorer])
                    if result[i][j].original_response:
                        score_list_i.append(result[i][j].scores[scorer])
                original_claim_scores[scorer].append(score_list_i)
                master_claim_scores[scorer].append(master_score_list_i)

            response_scores = self._aggregate_scores(original_claim_scores[scorer])
            self.scores_dict[scorer] = response_scores
        return original_claim_scores, master_claim_scores

    def _unpack_claims_data(self, result: List[List[Any]]) -> List[List[Dict[str, Any]]]:
        claims_data = []
        for i in range(len(result)):
            claims_data_i = []
            for j in range(len(result[i])):
                claim_dict_ij = result[i][j].dict()
                scores_dict_ij = claim_dict_ij.pop("scores")
                claim_dict_ij.update({k: s for k, s in scores_dict_ij.items() if k in self.scorers})
                claim_dict_ij["removed"] = False if not self.uad_result else self.uad_result["removed"][i][j]

                claims_data_i.append(claim_dict_ij)
            claims_data.append(claims_data_i)
        return claims_data



# Copyright 2025 CVS Health and/or one of its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Prompt templates and utilities for the UQLM library.
"""

from uqlm.utils.prompts.judge_prompts import TEMPLATE_TO_INSTRUCTION, TEMPLATE_TO_INSTRUCTION_WITH_EXPLANATIONS, SCORING_CONFIG, COMMON_INSTRUCTIONS, PROMPT_TEMPLATES, create_instruction
from uqlm.utils.prompts.decomposition import get_claim_breakdown_prompt, get_factoid_breakdown_template
from uqlm.utils.prompts.reconstruction import get_response_reconstruction_prompt
from uqlm.utils.prompts.entailment_prompts import get_entailment_prompt
from uqlm.utils.prompts.factscore_prompts import FACTSCORE_SYSTEM_PROMPT, SUBJECTIVE_SYSTEM_PROMPT

__all__ = ["TEMPLATE_TO_INSTRUCTION", "TEMPLATE_TO_INSTRUCTION_WITH_EXPLANATIONS", "SCORING_CONFIG", "COMMON_INSTRUCTIONS", "PROMPT_TEMPLATES", "create_instruction", "get_claim_breakdown_prompt", "get_response_reconstruction_prompt", "get_entailment_prompt", "FACTSCORE_SYSTEM_PROMPT", "SUBJECTIVE_SYSTEM_PROMPT"]



# Copyright 2025 CVS Health and/or one of its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def get_entailment_prompt(claim: str, source_text: str, style: str) -> str:
    """
    This is the entailment prompt from GraphUQ.

    Parameters
    ----------
    claim: str
        The claim to be evaluated.
    source_text: str
        The source text to be evaluated.

    Returns
    -------
    str
        The prompt template for evaluating the entailment of a claim and a source text.
    """

    entailment_prompt = None

    if style == "binary":  # this is modified version ofthe "edge construction prompt" from https://arxiv.org/pdf/2410.20783
        entailment_prompt = f"""
        Context: {source_text}
        Claim: {claim}
        Is the claim entailed by the context above?
        Answer Yes or No:
        """

    elif style == "p_true":
        entailment_prompt = f"""
        Source text:
        {source_text}
        Claim:
        {claim}

        Is the claim supported by the source text? 
        
        If so, answer "Yes". Otherwise, if the claim is simply neutral/subjective in regards to the source text or is contradicted by the source text, answer "No".
        
        Answer Yes or No:
        """

    elif style == "p_false":
        entailment_prompt = f"""
        Source text:
        {source_text}
        Claim:
        {claim}

        Is the claim contradicted by the source text? 
        
        If so, answer "Yes". Otherwise, if the claim is simply neutral/subjective in regards to the source text or is supported by the source text, answer "No".
        
        Answer Yes or No:
        """

    elif style == "p_neutral":
        entailment_prompt = f"""
        Source text:
        {source_text}
        Claim:
        {claim}

        Is the claim neutral to the source text? Meaning it is neither supported nor contradicted by the source text. 
        
        If the claim is neutral, answer "Yes". 
        
        If the claim is supported or contradicted by the source text, answer "No".

        Answer Yes or No:
        """

    elif style == "nli_classification":
        entailment_prompt = f"""
        
        Given the following premise and hypothesis, determine the natural language inference relationship.

        Premise: {source_text}

        Hypothesis: {claim}

        What is the relationship between the premise and hypothesis?
        - "entailment" if the hypothesis is supported by the premise
        - "contradiction" if the hypothesis contradicts the premise  
        - "neutral" if the hypothesis is neither supported nor contradicted by the premise

        Answer with only one word: entailment, contradiction, or neutral."""

    else:
        entailment_prompt = f"""

        You are a helpful assistant that can evaluate the entailment of a claim and a source text.

        You will be given a claim and a source text. You need to evaluate if the claim is entailed by the source text.

        You should return one of the following categorizations:

        true - if the claim is entailed by the source text.
        false - if the claim is not entailed by the source text.

        Example:

        Source text:
        Emory University has a strong athletics program, competing in the National Collegiate Athletic Association (NCAA) Division I Atlantic Coast Conference (ACC). The university's mascot is the Eagle.

        Claim:
        Emory University is part of the ACC.

        Categorization:
        true

        Only return the categorization label (true or false). Do not include any other text.

        Source text:
        {source_text}

        Claim:
        {claim}

        Categorization:
        """

    return entailment_prompt



# Copyright 2025 CVS Health and/or one of its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, List, Optional
from rich.progress import Progress
from langchain_core.language_models.chat_models import BaseChatModel
from uqlm.scorers.longform.baseclass.uncertainty import LongFormUQ
from uqlm.utils.results import UQResult
from uqlm.longform.luq.matched_unit import MatchedUnitScorer
from uqlm.longform.luq.unit_response import UnitResponseScorer

UNIT_RESPONSE_SCORERS = ["entailment", "noncontradiction", "contrasted_entailment"]
MATCHED_UNIT_SCORERS = UNIT_RESPONSE_SCORERS + ["bert_score", "cosine_sim"]


class LongTextUQ(LongFormUQ):
    def __init__(
        self,
        llm: Optional[BaseChatModel] = None,
        granularity: str = "claim",
        mode: str = "unit_response",
        scorers: Optional[List[str]] = None,
        aggregation: str = "mean",
        response_refinement: bool = False,
        claim_filtering_scorer: Optional[str] = None,
        claim_decomposition_llm: Optional[BaseChatModel] = None,
        device: Any = None,
        nli_model_name: str = "microsoft/deberta-large-mnli",
        system_prompt: str = "You are a helpful assistant.",
        max_calls_per_min: Optional[int] = None,
        sampling_temperature: float = 1.0,
        use_n_param: bool = False,
        max_length: int = 2000,
    ) -> None:
        """
        Class for Long-text Uncertainty Quantification (LUQ) scorers.

        Parameters
        ----------
        llm : langchain `BaseChatModel`, default=None
            A langchain llm `BaseChatModel`. User is responsible for specifying temperature and other
            relevant parameters to the constructor of their `llm` object.

        scorers : List[str], default=None
            Specifies which black box (consistency) scorers to include. List must be subset of ["entailment", "noncontradiction", "contrasted_entailment", "bert_score", "cosine_sim"].
            If None, defaults to ["entailment"].

        granularity : str, default="claim"
            Specifies whether to decompose and score at claim or sentence level granularity. Must be either "claim" or "sentence"

        aggregation : str, default="mean"
            Specifies how to aggregate claim/sentence-level scores to response-level scores. Must be one of 'min' or 'mean'.

        mode : str, default="unit_response"
            Specifies whether to implement unit-response (LUQ-style) scoring or matched-unit (LUQ-pair-style) scoring. Must be
            either "unit_response" or "matched_unit".

        response_refinement : bool, default=False
            Specifies whether to refine responses with uncertainty-aware decoding. This approach removes claims with confidence
            scores below the response_refinement_threshold and uses the claim_decomposition_llm to reconstruct the response from
            the retained claims. Only available for claim-level granularity. For more details, refer to
            Jiang et al., 2024: https://arxiv.org/abs/2410.20783

        claim_filtering_scorer : Optional[str], default=None
            specifies which scorer to use to filter claims if response_refinement is True. If not provided, defaults to the first
            element of self.scorers.

        claim_decomposition_llm : langchain `BaseChatModel`, default=None
            A langchain llm `BaseChatModel` to be used for decomposing responses into individual claims. Also used for claim refinement.
            If granularity="claim" and claim_decomposition_llm is None, the provided `llm` will be used for claim decomposition.

        device: str or torch.device input or torch.device object, default="cpu"
            Specifies the device that NLI model use for prediction. If None, detects and returns the best available PyTorch device.
            Prioritizes CUDA (NVIDIA GPU), then MPS (macOS), then CPU.

        nli_model_name : str, default="microsoft/deberta-large-mnli"
            Specifies which NLI model to use. Must be acceptable input to AutoTokenizer.from_pretrained() and
            AutoModelForSequenceClassification.from_pretrained()

        system_prompt : str or None, default="You are a helpful assistant."
            Optional argument for user to provide custom system prompt

        max_calls_per_min : int, default=None
            Specifies how many api calls to make per minute to avoid a rate limit error. By default, no
            limit is specified.

        sampling_temperature : float, default=1.0
            The 'temperature' parameter for llm model to generate sampled LLM responses. Must be greater than 0.

        use_n_param : bool, default=False
            Specifies whether to use `n` parameter for `BaseChatModel`. Not compatible with all
            `BaseChatModel` classes. If used, it speeds up the generation process substantially when num_responses > 1.

        max_length : int, default=2000
            Specifies the maximum allowed string length. Responses longer than this value will be truncated to
            avoid OutOfMemoryError
        """
        self.scorers = ["entailment"] if not scorers else scorers
        super().__init__(llm=llm, granularity=granularity, aggregation=aggregation, scorers=self.scorers, response_refinement=response_refinement, claim_filtering_scorer=claim_filtering_scorer, claim_decomposition_llm=claim_decomposition_llm, device=device, system_prompt=system_prompt, max_calls_per_min=max_calls_per_min, use_n_param=use_n_param)
        self.nli_model_name = nli_model_name
        self.mode = mode
        self.max_length = max_length
        self.sampling_temperature = sampling_temperature
        self._validate_scorers()
        self.prompts = None
        self.responses = None
        self.claim_sets = None
        self.sampled_responses = None
        self.sampled_claim_sets = None
        self.num_responses = None
        self.uad_result = {}

    async def generate_and_score(self, prompts: List[str], num_responses: int = 5, response_refinement_threshold: float = 1 / 3, show_progress_bars: Optional[bool] = True) -> UQResult:
        """
        Generate LLM responses, sampled LLM (candidate) responses, and compute confidence scores with specified scorers for the provided prompts.

        Parameters
        ----------
        prompts : list of str
            A list of input prompts for the model.

        num_responses : int, default=5
            The number of sampled responses used to compute consistency.

        response_refinement_threshold : float, default=1/3
            Threshold for uncertainty-aware filtering. Claims with confidence scores below this threshold are dropped from the
            refined response. Only used if response_refinement is True.

        show_progress_bars : bool, default=True
            If True, displays progress bars while generating and scoring responses

        Returns
        -------
        UQResult
            UQResult containing data (prompts, responses, and scores) and metadata
        """
        self.prompts = prompts
        self.num_responses = num_responses

        self._construct_progress_bar(show_progress_bars)
        self._display_generation_header(show_progress_bars)

        responses = await self.generate_original_responses(prompts=prompts, progress_bar=self.progress_bar)
        sampled_responses = await self.generate_candidate_responses(prompts=prompts, progress_bar=self.progress_bar, num_responses=self.num_responses)
        result = await self.score(responses=responses, sampled_responses=sampled_responses, response_refinement_threshold=response_refinement_threshold, show_progress_bars=show_progress_bars)
        return result

    async def score(self, responses: List[str], sampled_responses: List[List[str]], response_refinement_threshold: float = 1 / 3, show_progress_bars: Optional[bool] = True) -> UQResult:
        """
        Compute confidence scores with specified scorers on provided LLM responses. Should only be used if responses and sampled responses
        are already generated. Otherwise, use `generate_and_score`.

        Parameters
        ----------
        responses : list of str, default=None
            A list of model responses for the prompts.

        sampled_responses : list of list of str, default=None
            A list of lists of sampled LLM responses for each prompt. These will be used to compute consistency scores by comparing to
            the corresponding response from `responses`.

        response_refinement_threshold : float, default=1/3
            Threshold for uncertainty-aware filtering. Claims with confidence scores below this threshold are dropped from the
            refined response. Only used if response_refinement is True.

        show_progress_bars : bool, default=True
            If True, displays a progress bar while scoring responses

        Returns
        -------
        UQResult
            UQResult containing data (responses and scores) and metadata
        """
        self.responses = responses
        self.sampled_responses = sampled_responses
        self.num_responses = len(sampled_responses[0])
        self._construct_progress_bar(show_progress_bars)

        await self._decompose_responses(show_progress_bars)
        if self.mode == "matched_unit":
            await self._decompose_candidate_responses(show_progress_bars)
        self._display_scoring_header(show_progress_bars)

        self.scores_dict = self._score_from_decomposed(claim_sets=self.claim_sets, sampled_responses=self.sampled_responses, sampled_claim_sets=self.sampled_claim_sets, progress_bar=self.progress_bar)

        if self.response_refinement:
            self.uad_result = await self.uncertainty_aware_decode(claim_sets=self.claim_sets, claim_scores=self.claim_scores[self.uad_scorer], response_refinement_threshold=response_refinement_threshold, show_progress_bars=show_progress_bars)
        self._stop_progress_bar()
        self.progress_bar = None

        claims_data = []
        for i in range(len(self.claim_sets)):
            claim_i_data = []
            for j in range(len(self.claim_sets[i])):
                claims_dict = {self.granularity: self.claim_sets[i][j], "removed": False if not self.uad_result else self.uad_result["removed"][i][j]}
                for scorer in self.scorers:
                    claims_dict[scorer] = self.claim_scores[scorer][i][j]
                claim_i_data.append(claims_dict)
            claims_data.append(claim_i_data)

        self.scores_dict["claims_data"] = claims_data
        if "removed" in self.uad_result:
            del self.uad_result["removed"]

        return self._construct_result()

    def _score_from_decomposed(self, claim_sets: List[List[str]], sampled_responses: Optional[List[List[str]]] = None, sampled_claim_sets: Optional[List[List[List[str]]]] = None, progress_bar: Optional[Progress] = None) -> UQResult:
        """
        Compute confidence scores with specified scorers on provided LLM responses. Should only be used if responses and sampled responses
        are already generated. Otherwise, use `generate_and_score`.

        Parameters
        ----------
        claim_sets : list of list of strings
            List of original responses decomposed into lists of either claims or sentences

        sampled_responses : list of list of strings
            Candidate responses to be compared to the decomposed original responses

        sampled_claim_sets : list of list of list of strings
            Decomposed responses to be compared to the decomposed original responses

        Returns
        -------
        UQResult
            UQResult containing data (responses and scores) and metadata
        """
        if self.mode == "unit_response":
            self.claim_scores = self.unit_response_scorer.evaluate(claim_sets=self.claim_sets, sampled_responses=sampled_responses, progress_bar=progress_bar).to_dict()
        elif self.mode == "matched_unit":
            self.claim_scores = self.matched_unit_scorer.evaluate(claim_sets=self.claim_sets, sampled_claim_sets=self.sampled_claim_sets, progress_bar=progress_bar).to_dict()

        scores_dict = {}
        for scorer in self.scorers:
            scores_dict[scorer] = self._aggregate_scores(self.claim_scores[scorer])

        return scores_dict

    def _construct_result(self) -> Any:
        """Constructs UQResult object"""
        data = {"responses": self.responses, "sampled_responses": self.sampled_responses}
        if self.prompts:
            data["prompts"] = self.prompts
        # if self.claim_sets:
        #     data[self.granularity + "s"] = self.claim_sets
        data.update(self.scores_dict)
        data.update(self.uad_result)
        result = {"data": data, "metadata": {"mode": self.mode, "granularity": self.granularity, "aggregation": self.aggregation, "temperature": None if not self.llm else self.llm.temperature, "sampling_temperature": None if not self.sampling_temperature else self.sampling_temperature, "num_responses": self.num_responses, "response_refinement_threshold": self.response_refinement_threshold}}
        return UQResult(result)

    def _validate_scorers(self) -> None:
        """Validate scorers"""
        self.matched_unit_scorer = None
        self.unit_response_scorer = None
        if self.mode == "unit_response":
            if set(self.scorers) - set(UNIT_RESPONSE_SCORERS):
                raise ValueError(
                    f"""
                Invalid scorers: {set(self.scorers) - set(UNIT_RESPONSE_SCORERS)}. Must be subset of {UNIT_RESPONSE_SCORERS} when mode="unit_response"
                """
                )
            self.unit_response_scorer = UnitResponseScorer(nli_model_name=self.nli_model_name, device=self.device, max_length=self.max_length)
        elif self.mode == "matched_unit":
            self.matched_unit_scorer = MatchedUnitScorer(nli_model_name=self.nli_model_name, device=self.device, max_length=self.max_length)
            if set(self.scorers) - set(MATCHED_UNIT_SCORERS):
                raise ValueError(
                    f"""
                Invalid scorers: {set(self.scorers) - set(MATCHED_UNIT_SCORERS)}. Must be subset of {MATCHED_UNIT_SCORERS} when mode="matched_unit"
                """
                )
        else:
            raise ValueError(
                f"""
                Invalid mode: {self.mode}. Must be one of "unit_response", "matched_unit"
                """
            )



# Copyright 2025 CVS Health and/or one of its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch


def get_best_device() -> torch.device:
    """
    Detects and returns the best available PyTorch device.
    Prioritizes CUDA (NVIDIA GPU), then MPS (macOS), then CPU.

    Returns
    -------
    torch.device
        The best available device.

    Examples
    --------
    >>> device = get_best_device()
    >>> print(f"Using {device.type} device")
    """
    if torch.cuda.is_available():
        return torch.device("cuda")
    elif torch.backends.mps.is_available():
        return torch.device("mps")
    else:
        return torch.device("cpu")



# Copyright 2025 CVS Health and/or one of its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Any, List, Optional
import numpy as np
from langchain_core.language_models.chat_models import BaseChatModel
from uqlm.scorers.baseclass.uncertainty import UncertaintyQuantifier
from uqlm.longform.decomposition import ResponseDecomposer
from uqlm.longform.uad import UncertaintyAwareDecoder


class LongFormUQ(UncertaintyQuantifier):
    def __init__(
        self, llm: Any = None, scorers: Optional[List[str]] = None, granularity: str = "claim", aggregation: str = "mean", claim_decomposition_llm: Optional[BaseChatModel] = None, response_refinement: bool = False, claim_filtering_scorer: Optional[str] = None, device: Any = None, system_prompt: Optional[str] = None, max_calls_per_min: Optional[int] = None, use_n_param: bool = False
    ) -> None:
        """
        Parent class for uncertainty quantification of LLM responses

        Parameters
        ----------
        llm : BaseChatModel
            A langchain llm object to get passed to chain constructor. User is responsible for specifying
            temperature and other relevant parameters to the constructor of their `llm` object.

        scorers : List[str], default=None
            Specifies which black box (consistency) scorers to include.

        aggregation : str, default="mean"
            Specifies how to aggregate claim/sentence-level scores to response-level scores. Must be one of 'min' or 'mean'.

        granularity : str, default="claim"
            Specifies whether to decompose and score at claim or sentence level granularity. Must be either "claim" or "sentence"

        claim_decomposition_llm : langchain `BaseChatModel`, default=None
            A langchain llm `BaseChatModel` to be used for decomposing responses into individual claims. Also used for claim refinement.
            If granularity="claim" and claim_decomposition_llm is None, the provided `llm` will be used for claim decomposition.

        response_refinement : bool, default=False
            Specifies whether to refine responses with uncertainty-aware decoding. This approach removes claims with confidence
            scores below the response_refinement_threshold and uses the claim_decomposition_llm to reconstruct the response from
            the retained claims. Only available for claim-level granularity. For more details, refer to
            Jiang et al., 2024: https://arxiv.org/abs/2410.20783

        claim_filtering_scorer : Optional[str], default=None
            specifies which scorer to use to filter claims if response_refinement is True. If not provided, defaults to the first
            element of self.scorers.

        device: str or torch.device input or torch.device object, default="cpu"
            Specifies the device that NLI model use for prediction. Only applies to 'semantic_negentropy', 'noncontradiction'
            scorers. Pass a torch.device to leverage GPU.

        system_prompt : str, default=None
            Optional argument for user to provide custom system prompt. If prompts are list of strings and system_prompt is None,
            defaults to "You are a helpful assistant."

        max_calls_per_min : int, default=None
            Specifies how many api calls to make per minute to avoid a rate limit error. By default, no
            limit is specified.

        use_n_param : bool, default=False
            Specifies whether to use `n` parameter for `BaseChatModel`. Not compatible with all
            `BaseChatModel` classes. If used, it speeds up the generation process substantially when num_responses > 1.
        """
        super().__init__(llm=llm, device=device, system_prompt=system_prompt, max_calls_per_min=max_calls_per_min, use_n_param=use_n_param)
        self.claim_decomposition_llm = claim_decomposition_llm
        self.decomposer = ResponseDecomposer(claim_decomposition_llm=claim_decomposition_llm if claim_decomposition_llm else llm)
        self.granularity = granularity
        self.scorers = scorers
        self.aggregation = aggregation
        self.response_refinement = response_refinement
        self.claim_filtering_scorer = claim_filtering_scorer
        self.response_refinement_threshold = None
        if self.granularity not in ["sentence", "claim"]:
            raise ValueError(
                f"""
                Invalid granularity: {self.granularity}. Must be one of "sentence", "claim"
                """
            )
        if self.response_refinement:
            if self.granularity != "claim":
                raise ValueError("Uncertainty aware decoding is only possible with claim-level scoring. Please set response_refinement=False or granularity='claim'")
            self.reconstructor = UncertaintyAwareDecoder(reconstructor_llm=self.decomposer.claim_decomposition_llm)
            if not self.claim_filtering_scorer:
                print(f"claim_filtering_scorer is not specified for response_refinement. Defaulting to {self.scorers[0]}.")
                self.uad_scorer = self.scorers[0]
            elif self.claim_filtering_scorer not in self.scorers:
                print(f"claim_filtering_scorer is contained in list of scorers. Defaulting to {self.scorers[0]}.")
                self.uad_scorer = self.scorers[0]
            else:
                self.uad_scorer = self.claim_filtering_scorer

    async def uncertainty_aware_decode(self, claim_sets: List[List[str]], claim_scores: List[List[float]], response_refinement_threshold: float = 1 / 3, show_progress_bars: Optional[bool] = True) -> List[str]:
        """
        Parameters
        ----------
        claim_sets : List[List[str]]
            List of original responses decomposed into lists of claims

        claim_scores : List[List[float]]
            List of lists of claim-level confidence scores to be used for uncertainty-aware filtering

        response_refinement_threshold : float, default=1/3
            Threshold for uncertainty-aware filtering. Claims with confidence scores below this threshold are dropped from the
            refined response. Only used if response_refinement is True.

        progress_bar : rich.progress.Progress, default=None
            If provided, displays a progress bar while scoring responses
        """
        self._construct_progress_bar(show_progress_bars)
        self._display_reconstruction_header(show_progress_bars)
        self.response_refinement_threshold = response_refinement_threshold
        uad_result = await self.reconstructor.reconstruct_responses(claim_sets=claim_sets, claim_scores=claim_scores, responses=self.responses, threshold=response_refinement_threshold, progress_bar=self.progress_bar)
        self._stop_progress_bar()
        self.progress_bar = None

        for scorer in self.scorers:
            filtered_claim_scores = []
            for i in range(len(claim_sets)):
                filtered_claim_scores_i = []
                for j in range(len(claim_sets[i])):
                    if not uad_result["removed"][i][j]:
                        filtered_claim_scores_i.append(self.claim_scores[scorer][i][j])
                filtered_claim_scores.append(filtered_claim_scores_i)

            uad_result["refined_" + scorer] = self._aggregate_scores(filtered_claim_scores)

        return uad_result

    async def _decompose_responses(self, show_progress_bars) -> None:
        """Decompose original responses into claims or sentences"""
        self._display_decomposition_header(show_progress_bars)
        if self.granularity == "sentence":
            self.claim_sets = self.decomposer.decompose_sentences(responses=self.responses, progress_bar=self.progress_bar)
        elif self.granularity == "claim":
            self.claim_sets = await self.decomposer.decompose_claims(responses=self.responses, progress_bar=self.progress_bar)

    async def _decompose_candidate_responses(self, show_progress_bars) -> None:
        """Display header and decompose responses"""
        if self.granularity == "sentence":
            self.sampled_claim_sets = self.decomposer.decompose_candidate_sentences(sampled_responses=self.sampled_responses, progress_bar=self.progress_bar)
        elif self.granularity == "claim":
            self.sampled_claim_sets = await self.decomposer.decompose_candidate_claims(sampled_responses=self.sampled_responses, progress_bar=self.progress_bar)

    def _aggregate_scores(self, claim_scores: List[List[float]]) -> List[float]:
        """Aggregate claim scores to response level scores"""
        if self.aggregation == "mean":
            return [np.mean(cs) for cs in claim_scores]
        elif self.aggregation == "min":
            return [np.min(cs) for cs in claim_scores]

    def _display_decomposition_header(self, show_progress_bars: bool) -> None:
        """Displays decomposition header"""
        if show_progress_bars:
            self.progress_bar.start()
            self.progress_bar.add_task("")
            self.progress_bar.add_task("✂️ Decomposition")

    def _display_reconstruction_header(self, show_progress_bars: bool) -> None:
        """Displays decomposition header"""
        if show_progress_bars:
            self.progress_bar.start()
            self.progress_bar.add_task("")
            self.progress_bar.add_task("✅️ Refinement")



# Copyright 2025 CVS Health and/or one of its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional
from rich.progress import Progress
from uqlm.utils.prompts.claims_prompts import get_claim_dedup_prompt
from uqlm.utils.response_generator import ResponseGenerator
from langchain_core.language_models.chat_models import BaseChatModel
import re


class ClaimMerger:
    def __init__(self, claim_merging_llm: BaseChatModel) -> None:
        self.rg = ResponseGenerator(llm=claim_merging_llm)

    async def merge_claims(self, original_claim_sets: List[List[str]], sampled_claim_sets: List[List[List[str]]], progress_bar: Optional[Progress] = None) -> List[List[str]]:
        """Process claim deduplication for response sets.
        Leverages ResponseGenerator's ability to handle multiple prompts at once
        by collecting dedup prompts and making batch calls.
        If sampled_claim_sets contains only empty lists and entailment_score_sets is provided,
        infers master claims from entailment_score_sets keys. Otherwise returns original_claim_sets.
        """
        num_response_sets = len(original_claim_sets)
        num_samples = len(sampled_claim_sets[0]) * num_response_sets

        self.progress_task = None
        if progress_bar:
            self.progress_task = progress_bar.add_task("  - Deduplicating claims across candidate responses...", total=num_samples)

        self.master_claim_sets = [original_claim_sets[i] for i in range(num_response_sets)]
        for iteration in range(num_samples):
            prompts, prompt_metadata = self._construct_merging_prompts(sampled_claim_sets=sampled_claim_sets, iteration=iteration)

            # Batch call for this iteration across all response sets
            if prompts:
                result = await self.rg.generate_responses(prompts=prompts)
                responses = result["data"]["response"]
            else:
                responses = []

            self._process_claim_merging_generations(responses, prompt_metadata, progress_bar)

        return self.master_claim_sets

    def _process_claim_merging_generations(self, responses, prompt_metadata, progress_bar) -> None:
        # Update master_claim_sets with results
        response_idx = 0
        for i, has_prompt, current_master, sampled_claims in prompt_metadata:
            if has_prompt and response_idx < len(responses):
                response_text = responses[response_idx]
                new_claims = re.findall(r"^\s*-\s*(.+)", response_text, re.MULTILINE)

                if new_claims:
                    self.master_claim_sets[i] = current_master + new_claims

                response_idx += 1

        if progress_bar and self.progress_task is not None:
            progress_bar.update(self.progress_task, advance=len(responses))

    def _construct_merging_prompts(self, sampled_claim_sets, iteration):
        prompts = []
        prompt_metadata = []  # (response_set_idx, has_prompt)

        for i in range(len(self.master_claim_sets)):
            if iteration < len(sampled_claim_sets[i]):
                sampled_claims = sampled_claim_sets[i][iteration]
                unique_sampled_claims = list(set(sampled_claims) - set(self.master_claim_sets[i]))

                if unique_sampled_claims:
                    prompts.append(get_claim_dedup_prompt(self.master_claim_sets[i], unique_sampled_claims))
                    prompt_metadata.append((i, True, self.master_claim_sets[i], sampled_claims))
                else:
                    prompt_metadata.append((i, False, self.master_claim_sets[i], sampled_claims))
            else:
                prompt_metadata.append((i, False, self.master_claim_sets[i], []))

        return prompts, prompt_metadata



# Copyright 2025 CVS Health and/or one of its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import time
from typing import List, Optional
from uqlm.utils.prompts import get_response_reconstruction_prompt
from rich.progress import Progress
from langchain_core.language_models.chat_models import BaseChatModel


class UncertaintyAwareDecoder:
    def __init__(self, reconstructor_llm: BaseChatModel) -> None:
        """
        Class for decomposing responses into individual claims or sentences. This class is used as an intermediate
        step for longform UQ methods.

        Parameters
        ----------
        reconstructor_llm : langchain `BaseChatModel`, default=None
            A langchain llm `BaseChatModel`. User is responsible for specifying temperature and other
            relevant parameters to the constructor of their `llm` object.
        """
        self.reconstructor_llm = reconstructor_llm
        self.reconstruction_template = get_response_reconstruction_prompt

    async def reconstruct_responses(self, claim_sets: List[List[str]], claim_scores: List[List[float]], responses: Optional[List[str]] = None, threshold: float = 1 / 3, progress_bar: Optional[Progress] = None) -> List[str]:
        """
        Parameters
        ----------
        claim_sets : List[List[str]]
            List of original responses decomposed into lists of claims

        claim_scores : List[List[float]]
            List of lists of claim-level confidence scores to be used for uncertainty-aware filtering

        threshold : float, default=1/3
            Threshold used for uncertainty-aware filtering

        progress_bar : rich.progress.Progress, default=None
            If provided, displays a progress bar while scoring responses
        """
        if not responses:
            responses = [None] * len(claim_sets)

        filtered_claim_scores, filtered_claim_sets, remove_indicators = [], [], []
        for i in range(len(claim_sets)):
            filtered_claim_scores_i, filtered_claim_set_i, remove_indicators_i = [], [], []
            for j in range(len(claim_scores[i])):
                if claim_scores[i][j] > threshold:
                    filtered_claim_scores_i.append(claim_scores[i][j])
                    filtered_claim_set_i.append(claim_sets[i][j])
                    remove_indicators_i.append(False)
                else:
                    remove_indicators_i.append(True)
            filtered_claim_scores.append(filtered_claim_scores_i)
            filtered_claim_sets.append(filtered_claim_set_i)
            remove_indicators.append(remove_indicators_i)

        if progress_bar:
            self.progress_task = progress_bar.add_task(" - Reconstructing responses with high-confidence claims...", total=len(claim_sets))
        tasks = [self._reconstruct_single_response(claim_set=filtered_claim_sets[i], response=responses[i], progress_bar=progress_bar) for i in range(len(claim_sets))]
        reconstructed_responses = await asyncio.gather(*tasks)
        time.sleep(0.1)
        return {"refined_responses": reconstructed_responses, "removed": remove_indicators}

    async def _reconstruct_single_response(self, claim_set: List[str], response: Optional[str] = None, progress_bar: Optional[Progress] = None) -> str:
        """Decompose single response into claims using LLM and extract claims from the result"""
        if claim_set:
            generation = await self.reconstructor_llm.ainvoke(self.reconstruction_template(claim_set))
            reconstructed_response = generation.content
        else:
            reconstructed_response = response
        if progress_bar:
            progress_bar.update(self.progress_task, advance=1)
        return reconstructed_response
